@@ -3,9 +3,10 @@
 #![allow(clippy::integer_arithmetic)]
 use {
     common::{
-        copy_blocks, create_custom_leader_schedule, last_vote_in_tower, ms_for_n_slots,
-        open_blockstore, restore_tower, run_cluster_partition, run_kill_partition_switch_threshold,
-        test_faulty_node, wait_for_last_vote_in_tower_to_land_in_ledger, RUST_LOG_FILTER,
+        copy_blocks, create_custom_leader_schedule_with_random_keys, last_vote_in_tower,
+        ms_for_n_slots, open_blockstore, restore_tower, run_cluster_partition,
+        run_kill_partition_switch_threshold, test_faulty_node,
+        wait_for_last_vote_in_tower_to_land_in_ledger, RUST_LOG_FILTER,
     },
     log::*,
     serial_test::serial,
@@ -279,7 +280,7 @@ fn test_kill_heaviest_partition() {
     // 4) Check for recovery
     let num_slots_per_validator = 8;
     let partitions: [Vec<usize>; 4] = [vec![11], vec![10], vec![10], vec![10]];
-    let (leader_schedule, validator_keys) = create_custom_leader_schedule(&[
+    let (leader_schedule, validator_keys) = create_custom_leader_schedule_with_random_keys(&[
         num_slots_per_validator * (partitions.len() - 1),
         num_slots_per_validator,
         num_slots_per_validator,
@@ -876,46 +877,27 @@ fn find_latest_replayed_slot_from_ledger(
             latest_slot = new_latest_slot;
             info!("Checking latest_slot {}", latest_slot);
             // Wait for the slot to be fully received by the validator
-            let entries;
             loop {
                 info!("Waiting for slot {} to be full", latest_slot);
                 if blockstore.is_full(latest_slot) {
-                    entries = blockstore.get_slot_entries(latest_slot, 0).unwrap();
-                    assert!(!entries.is_empty());
                     break;
                 } else {
                     sleep(Duration::from_millis(50));
                     blockstore = open_blockstore(ledger_path);
                 }
             }
-            // Check the slot has been replayed
-            let non_tick_entry = entries.into_iter().find(|e| !e.transactions.is_empty());
-            if let Some(non_tick_entry) = non_tick_entry {
-                // Wait for the slot to be replayed
-                loop {
-                    info!("Waiting for slot {} to be replayed", latest_slot);
-                    if !blockstore
-                        .map_transactions_to_statuses(
-                            latest_slot,
-                            non_tick_entry.transactions.clone().into_iter(),
-                        )
-                        .unwrap()
-                        .is_empty()
-                    {
-                        return (
-                            latest_slot,
-                            AncestorIterator::new(latest_slot, &blockstore).collect(),
-                        );
-                    } else {
-                        sleep(Duration::from_millis(50));
-                        blockstore = open_blockstore(ledger_path);
-                    }
+            // Wait for the slot to be replayed
+            loop {
+                info!("Waiting for slot {} to be replayed", latest_slot);
+                if blockstore.get_bank_hash(latest_slot).is_some() {
+                    return (
+                        latest_slot,
+                        AncestorIterator::new(latest_slot, &blockstore).collect(),
+                    );
+                } else {
+                    sleep(Duration::from_millis(50));
+                    blockstore = open_blockstore(ledger_path);
                 }
-            } else {
-                info!(
-                    "No transactions in slot {}, can't tell if it was replayed",
-                    latest_slot
-                );
             }
         }
         sleep(Duration::from_millis(50));
